import { toPointer } from '@ui-schema/json-pointer/toPointer';
import { ValidatorOutput } from '@ui-schema/ui-schema/ValidatorOutput';
import { List, is } from 'immutable';
export const ERROR_DUPLICATE_ITEMS = 'duplicate-items';
export const ERROR_NOT_FOUND_CONTAINS = 'not-found-contains';
export const ERROR_MIN_CONTAINS = 'min-contains';
export const ERROR_MAX_CONTAINS = 'max-contains';
export const ERROR_ADDITIONAL_ITEMS = 'additional-items';
const findDuplicates = (arr, isEqual) => {
  const arrInterop = Array.isArray(arr) ? {
    size: arr.length,
    get: i => arr[i]
  } : arr;
  const seen = [];
  for (let i = 0; i < arrInterop.size; i++) {
    const val = arrInterop.get(i);
    const existing = seen.find(entry => isEqual(entry.value, val));
    if (existing) {
      existing.indexes.push(i);
    } else {
      seen.push({
        value: val,
        indexes: [i]
      });
    }
  }
  return seen.filter(entry => entry.indexes.length > 1);
};
export const validateUniqueItems = (schema, value, params) => {
  const uniqueItems = schema.get('uniqueItems');
  if (!uniqueItems) {
    return true;
  }
  const duplicates = findDuplicates(value, is);
  if (duplicates.length > 0 && params) {
    params.output.addError({
      error: ERROR_DUPLICATE_ITEMS,
      keywordLocation: toPointer([...params.keywordLocation, 'uniqueItems']),
      instanceLocation: toPointer(params.instanceLocation),
      context: {
        duplicates: duplicates.map(d => ({
          value: d.value,
          indexes: d.indexes
        }))
      }
    });
  }
  return duplicates.length === 0;
};
export const validateItemsAgainstSchema = (itemsToValidate, schemaToValidate, params) => {
  let found = 0;
  let i = 0;
  for (const val of itemsToValidate) {
    const result = params.validate(schemaToValidate, val, {
      ...params,
      instanceLocation: [...params.instanceLocation, i],
      instanceKey: i,
      output: new ValidatorOutput(),
      context: {}
    });
    if (result.valid) {
      found++;
    }
    i++;
  }
  return {
    found
  };
};
const normalizeArraySchema = schema => {
  const itemsTmp = schema.get('items');
  const prefixItemsTmp = schema.get('prefixItems');
  const additionalItemsTmp = schema.get('additionalItems');
  let prefixItems = undefined;
  let items = undefined;
  let additionalItemsAllowed = true;
  if (prefixItemsTmp !== undefined && List.isList(prefixItemsTmp)) {
    prefixItems = prefixItemsTmp;
    if (itemsTmp === false) {
      additionalItemsAllowed = false;
    } else if (typeof itemsTmp === 'object' && itemsTmp !== null) {
      items = itemsTmp;
    }
  } else if (List.isList(itemsTmp)) {
    prefixItems = itemsTmp;
    if (additionalItemsTmp === false) {
      additionalItemsAllowed = false;
    } else if (typeof additionalItemsTmp === 'object' && additionalItemsTmp !== null) {
      items = additionalItemsTmp;
    }
  } else if (typeof itemsTmp === 'object' && itemsTmp !== null) {
    items = itemsTmp;
  } else if (itemsTmp === false) {
    additionalItemsAllowed = false;
  }
  return {
    prefixItems,
    items,
    additionalItemsAllowed
  };
};
export const validateItems = (schema, value, params) => {
  const {
    prefixItems,
    items,
    additionalItemsAllowed
  } = normalizeArraySchema(schema);
  const prefixItemsCount = prefixItems?.size || 0;
  let i = 0;
  for (const val of value) {
    let itemSchema = undefined;
    let keyword;
    if (i < prefixItemsCount) {
      itemSchema = prefixItems.get(i);
      keyword = schema.has('prefixItems') ? 'prefixItems' : 'items';
    } else {
      if (!additionalItemsAllowed) {
        params.output.addError({
          error: ERROR_ADDITIONAL_ITEMS,
          keywordLocation: toPointer([...params.keywordLocation, schema.get('items') === false ? 'items' : 'additionalItems']),
          instanceLocation: toPointer([...params.instanceLocation, i])
        });
        i++;
        continue;
      }
      itemSchema = items;
      keyword = 'items';
    }
    if (itemSchema) {
      params.validate(itemSchema, val, {
        ...params,
        keywordLocation: prefixItems?.has(i) ? [...params.keywordLocation, keyword, i] : [...params.keywordLocation, keyword],
        instanceLocation: [...params.instanceLocation, i],
        instanceKey: i
      });
    }
    i++;
  }
};
export const validateContains = (schema, value, params) => {
  const containsSchema = schema.get('contains');
  if (typeof containsSchema !== 'object' || containsSchema === null) return;
  const minContains = schema.get('minContains') ?? 1;
  const maxContains = schema.get('maxContains');
  const {
    found
  } = validateItemsAgainstSchema(value, containsSchema, params);
  if (found < minContains) {
    const isDefaultMinContains = minContains === 1 && !schema.has('minContains');
    params.output.addError({
      error: isDefaultMinContains ? ERROR_NOT_FOUND_CONTAINS : ERROR_MIN_CONTAINS,
      keywordLocation: toPointer([...params.keywordLocation, isDefaultMinContains ? 'contains' : 'minContains']),
      instanceLocation: toPointer(params.instanceLocation),
      context: {
        minContains,
        found
      }
    });
  }
  if (typeof maxContains === 'number' && found > maxContains) {
    params.output.addError({
      error: ERROR_MAX_CONTAINS,
      keywordLocation: toPointer([...params.keywordLocation, 'maxContains']),
      instanceLocation: toPointer(params.instanceLocation),
      context: {
        maxContains,
        found
      }
    });
  }
};
export const arrayValidator = {
  types: ['array'],
  validate: (schema, value, params) => {
    if (!(List.isList(value) || Array.isArray(value))) {
      return;
    }
    if (schema.get('uniqueItems')) {
      validateUniqueItems(schema, value, params);
    }
    if (schema.has('contains')) {
      validateContains(schema, value, params);
    }
    if (params.recursive) {
      validateItems(schema, value, params);
    }
  }
};